
-- import lib/List.agda
-- import lib/Bool.agda

-- Predefined sets --------------------------------------------------------

A : Set
L = List A

-- Ordering ---------------------------------------------------------------

infix 8 <=, >=, <, >
(<=) : A -> A -> Bool

(>=) : A -> A -> Bool
x >= y = y <= x

(<) : A -> A -> Bool
x < y = not (y <= x)

(>) : A -> A -> Bool
x > y = y < x

leq_trans : (x,y,z:A) -> T (x <= y) /\ T (y <= z) ==> T (x <= z)
leq_total : (x,y:A) -> T (x < y) ==> T (x <= y)

-- Merge ------------------------------------------------------------------

merge : (xs,ys : L) -> L
merge nil ys = ys
merge xs nil = xs
merge (x::xs) (y::ys) = if L (x <= y) (x :: merge xs (y::ys))
				    (y :: merge (x::xs) ys)

-- Merge induction --------------------------------------------------------

mergeInd : (P : L -> L -> Prop)   ->
	    ((ys : L) -> P nil ys) ->
	    ((xs : L) -> P xs nil) ->
	    ((x : A) -> (xs : L) -> (y : A) -> (ys : L) ->
	      T (x <= y) -> P xs (y::ys) -> P (x::xs) (y::ys)) ->
	    ((x : A) -> (xs : L) -> (y : A) -> (ys : L) ->
	      T (x > y) -> P (x::xs) ys -> P (x::xs) (y::ys))  ->
	    (xs, ys : L) -> P xs ys
mergeInd P b1 b2 s1 s2 nil ys = b1 ys
mergeInd P b1 b2 s1 s2 xs nil = b2 xs
mergeInd P b1 b2 s1 s2 (x::xs) (y::ys) =
  fol-plugin( trueOrFalse, s1, s2
	    , mergeInd P b1 b2 s1 s2 xs (y::ys)
	    , mergeInd P b1 b2 s1 s2 (x::xs) ys
	    )

-- << ---------------------------------------------------------------------

infix 8 <<
(<<) : A -> L -> Bool
x << nil    = true
x << (y::ys) = x <= y && x << ys

-- Ordered lists ----------------------------------------------------------

ordered : (xs : L) -> Bool
ordered nil = true
ordered (x::xs) = x << xs && ordered xs

-- Lemma 1:: Transitivity of << --------------------------------------------

Lem1 : (xs : L) -> Prop
Lem1 xs = (x,y:A) -> T (x <= y && y << xs) ==> T (x << xs)

lem1 : (xs : L) -> Lem1 xs
lem1 = listInd A Lem1
	  fol-plugin()
	  fol-plugin((<<), leq_trans, andT)

-- Lemma 2:: x <= y /\ ordered (y::xs) ==> ordered (x::xs) -------------------

Lem2 : (xs : L) -> Prop
Lem2 xs = (x,y:A) -> T (x <= y && ordered (y::xs)) -> T (ordered (x::xs))

lem2 : (xs : L) -> Lem2 xs
lem2 = fol-plugin(lem1, andT)

-- Lemma 3:: Merge preserves order under cons ------------------------------
Lem3 : (xs, ys : L) -> Prop
Lem3 xs ys = (z : A) ->
	      T (ordered (z::xs) && ordered (z::ys)) ==>
	      T (ordered (z :: merge xs ys))

lem3 = mergeInd Lem3
	fol-plugin ()
	fol-plugin (andT)
	(\x xs y ys p h ->
	  fol-plugin ( merge, andT, if, ordered, (<<), lem2, p, h x )
	)
	assert-plugin ()

{-
lem3 : (xs, ys : L) -> Lem3 xs ys
lem3 = listInd A
	  (\xs -> (ys:L) -> Lem3 xs ys)
	  fol-plugin ()
	  (\x xs ih1 ->
	    listInd A (Lem3 (x::xs))
	      fol-plugin (andT)
	      (\y ys ih2 ->
		  let case1 : T (x <= y) ==> Lem3 (x::xs) (y::ys)
		      case1 = assert-plugin ()
		      case2 : T (x > y) ==> Lem3 (x::xs) (y::ys)
		      case2 = assert-plugin()
		  in  fol-plugin (case1, case2, trueOrFalse)
	      )
	  )
-}
