-- Almost identical to Queue.agda 
-- List-induction is defined with implication instead of (->) in the cons case

-- # import Property

-- Booleans ---------------------------------------------------------------

Bool : Set

(=-=) : Bool -> Bool -> Prop
(=-=) a b = Eq Bool a b

true : Bool
false : Bool

true_false  : Not (true =-= false)

if	    : (A:Set) -> Bool -> A -> A -> A
if_true	    : (A:Set) -> (x,y:A) -> Eq A (if A true x y) x
if_false    : (A:Set) -> (x,y:A) -> Eq A (if A false x y) y

-- Lists ------------------------------------------------------------------

A  : Set
As : Set

(-=-) : A -> A -> Prop
(-=-) x y = Eq A x y

infix 9 ==, -=-, =-=
(==) : As -> As -> Prop
(==) xs ys = Eq As xs ys

nil : As
(::) : A -> As -> As
infixr 13 ::

nil_cons : (x:A) -> (xs:As) -> Not (nil == x::xs)

listInd	:  (P	    : As -> Prop) ->
	    (P_nil  : P nil) ->
	    (P_cons : (x:A) -> (xs:As) -> P xs ==> P (x::xs)) ->
	    (xs	    : As) ->
	    P xs

listCase : (P	    : As -> Prop) ->
	    (P_nil  : P nil) ->
	    (P_cons : (x:A) -> (xs:As) -> P (x::xs)) ->
	    (xs	    : As) ->
	    P xs
listCase = \P P_nil P_cons xs -> listInd P P_nil (\x xs' _ -> P_cons x xs') xs

-- List functions ---------------------------------------------------------

-- head
head : As -> A
head_cons : (x:A) -> (xs:As) -> head (x::xs) -=- x

-- tail
tail : As -> As
tail_cons : (x:A) -> (xs:As) -> tail (x::xs) == xs

-- null
null : As -> Bool
null_nil : null nil =-= true
null_cons : (x:A) -> (xs:As) -> null (x::xs) =-= false

null_true   : (xs:As) -> null xs =-= true ==> xs == nil
{-
null_false  : (xs:As) -> null xs =-= false ==>
				    Exists A  (\y ->
				    Exists As (\ys ->
					xs == y::ys
				    ) )
-}

-- (++)
infixr 11 ++
(++) : As -> As -> As

app_nil	 : (ys:As)		    -> nil  ++ ys == ys
app_cons : (x:A) -> (xs,ys:As)   -> x::xs ++ ys == x :: (xs ++ ys)

-- Properties of (++) ----

-- Associativity
AppAssoc : (xs:As) -> Prop
AppAssoc xs = (ys,zs:As) -> xs ++ (ys ++ zs) == (xs ++ ys) ++ zs

app_assoc : (xs:As) -> AppAssoc xs
app_assoc = listInd AppAssoc fol-plugin(app_nil) fol-plugin(app_cons)

-- xs ++ [] == xs
AppNil2 : (xs:As) -> Prop
AppNil2 xs = xs ++ nil == xs

app_nil2 : (xs:As) -> AppNil2 xs
app_nil2 = listInd AppNil2 fol-plugin(app_nil) fol-plugin(app_cons)

-- Reverse
reverse	: As -> As

reverse_nil  : reverse nil == nil
reverse_cons : (x:A) -> (xs:As) -> reverse (x::xs) == reverse xs ++ x::nil

-- Properties of reverse

-- reverse (xs ++ ys) == reverse ys ++ reverse xs
ReverseApp : (xs:As) -> Prop
ReverseApp xs = (ys:As) -> (reverse (xs ++ ys) == reverse ys ++ reverse xs)

reverse_app : (xs:As) -> ReverseApp xs
reverse_app = listInd	ReverseApp
			fol-plugin(reverse_nil, app_nil, app_nil2)
			fol-plugin(reverse_cons, app_cons, app_assoc)

-- Pairs ------------------------------------------------------------------

As2 : Set

pair : As -> As -> As2

fst : As2 -> As
fst_pair : (xs,ys:As) -> fst (pair xs ys) == xs

snd : As2 -> As
snd_pair : (xs,ys:As) -> snd (pair xs ys) == ys

-- Queue specification ----------------------------------------------------

Q = As

empty : Q
empty = nil

add : A -> Q -> Q
add x q = q ++ x::nil

isEmpty : Q -> Bool
isEmpty = null

front : Q -> A
front = head

remove : Q -> As
remove = tail

-- Queue implementation ---------------------------------------------------

QI = As2

flipQ : QI -> QI
flipQ q =   let f = fst q
		b = snd q
	    in	if QI	(null f)
			(pair (reverse b) nil)
			q

emptyI : QI
emptyI = pair nil nil

addI : A -> QI -> QI
addI x q =  let	f = fst q
		b = snd q
	    in  flipQ (pair f (x::b))

isEmptyI : QI -> Bool
isEmptyI q = null (fst q)

frontI : QI -> A
frontI q = head (fst q)

removeI : QI -> QI
removeI q = pair (tail (fst q)) (snd q)

retrieve : QI -> Q
retrieve q = fst q ++ reverse (snd q)

-- Queue properties -------------------------------------------------------

-- Our invariant
invariant : QI -> Prop
invariant q =
    let	f = fst q
	b = snd q
    in	null b =-= true \/ null f =-= false

-- invariant ([],b) ==> b == []
prop_inv : (b:As) -> invariant (pair nil b) ==> b == nil
prop_inv = fol-plugin (null_nil, null_true, true_false, fst_pair, snd_pair)

-- emptyI == empty
prop_empty : retrieve emptyI == empty
prop_empty = fol-plugin (reverse_nil, app_nil, null_nil, fst_pair, snd_pair)

-- emptyI establishes the invariant
prop_inv_empty : invariant emptyI
prop_inv_empty = fol-plugin (null_nil, fst_pair, snd_pair)

-- addI == add
PropAdd : (f:As) -> Prop
PropAdd f =
    (b:As) -> (x:A) ->
	let q = pair f b
	in  retrieve (addI x q) == add x (retrieve q)

prop_add : (f:As) -> PropAdd f
prop_add = listCase PropAdd
		fol-plugin
		    ( reverse_app, app_nil2, reverse_nil, reverse_cons
		    , fst_pair, snd_pair, if_true QI, null_nil
		    )
		fol-plugin
		    ( app_assoc, reverse_cons, fst_pair, snd_pair
		    , if_false QI, null_cons
		    )

-- addI preserves the invariant
PropInvAdd : (f:As) -> Prop
PropInvAdd f =
    (b:As) -> (x:A) ->
	let q = pair f b
	in  invariant q ==> invariant (addI x q)

prop_inv_add : (f:As) -> PropInvAdd f
prop_inv_add = listCase PropInvAdd
		    fol-plugin (null_nil, null_cons, fst_pair, snd_pair, if_true QI)
		    fol-plugin (null_cons, fst_pair, snd_pair, if_false QI)

-- isEmptyI == isEmpty
PropIsEmpty : (f:As) -> Prop
PropIsEmpty f =
    (b:As) ->
	let q = pair f b
	in  invariant q ==> isEmptyI q =-= isEmpty (retrieve q)

prop_isEmpty : (f:As) -> PropIsEmpty f
prop_isEmpty = listCase PropIsEmpty
		    fol-plugin	( null_nil, fst_pair, snd_pair, app_nil
				, reverse_nil, null_true, true_false
				)
		    fol-plugin	( null_cons, fst_pair, app_cons )

-- frontI == front
PropFront : (x:A) -> (f:As) -> Prop
PropFront x f =
    (b:As) ->
	let q = pair (x::f) b
	in  invariant q ==> frontI q -=- front (retrieve q)

prop_front : (x:A) -> (f:As) -> PropFront x f
prop_front = fol-plugin ( fst_pair, app_cons, head_cons )

-- removeI == remove
PropRemove : (x:A) -> (f:As) -> Prop
PropRemove x f =
    (b:As) ->
	let q = pair (x::f) b
	in  invariant q ==> retrieve (removeI q) == remove (retrieve q)

prop_remove : (x:A) -> (f:As) -> PropRemove x f
prop_remove = fol-plugin ( fst_pair, snd_pair, app_cons, tail_cons )

