
-- # import prelude
-- # import list
-- # import id
-- # import prop
-- # import eqchain

data PartialBijection (A, B : Set) : Set where
  pBij : (xs   : List (A × B)) ->
	 (injL : (x,y : A) -> (z : B) ->
		  Elem (pair x z) xs -> Elem (pair y z) xs ->
		  x == y
	 ) ->
	 (injR : (x : A) -> (y,z : B) ->
		 Elem (pair x y) xs -> Elem (pair x z) xs ->
		 y == z
	 ) -> PartialBijection A B

elements (A, B : Set) : PartialBijection A B -> List (A × B)
elements (pBij xs _ _) = xs

∅ (A, B : Set) : PartialBijection A B
∅ = pBij [] 
	 (\x y z ⊥ -> elemEmpty ⊥)
	 (\x y z ⊥ -> elemEmpty ⊥)

(∈) (A, B : Set) : A × B -> PartialBijection A B -> Set
p ∈ R = Elem p (elements R)

(⊆) (A,B:Set) : PartialBijection A B -> PartialBijection A B -> Set
Φ ⊆ Ψ = (z : _ × _) -> z ∈ Φ -> z ∈ Ψ

unitBijection (A, B : Set) : A × B -> PartialBijection A B
unitBijection p = pBij [p] (\x y z exz eyz ->
			      andE1 $ injPair $
				pair x z
				=-= p	      `since` elemSingleton exz
				=== pair y z  `since` sym (elemSingleton eyz)
			   )
			   (\x y z exy exz ->
			      andE2 $ injPair $
				pair x y
				=-= p	      `since` elemSingleton exy
				=== pair x z  `since` sym (elemSingleton exz)
			   )

unionBijection (A, B : Set) :
    (Φ1, Φ2 : PartialBijection A B) ->
    PartialBijection A B

-- When extending we can take stuff away. Extends Ψ Φ z doesn't mean that z is in
-- Ψ, it just means that it might be. Nor does it mean that Φ ⊆ Ψ. More precisely:
-- Extends Ψ Φ z ≡ Ψ ⊆ Φ ∪ {z}. This is to allow z to override mappings in Φ.
Extends (A,B:Set) : PartialBijection A B -> PartialBijection A B -> A × B -> Set
Extends {A} {B} Ψ Φ x =
  (z : A × B) -> z ∈ Ψ -> z == x \/ z ∈ Φ

-- IsIdentitity Φ ≡ Φ ⊆ Id
IsIdentity (A:Set) : PartialBijection A A -> Set
IsIdentity {A} Φ = (x, y : A) -> pair x y ∈ Φ -> x == y

unitIdentity (A:Set)(z:A) : IsIdentity (unitBijection (pair z z))
unitIdentity x y exy with elemSingleton exy =>
	     z z exy refl = refl

