{-# OPTIONS -cpp #-}

module IO.Pipes where

import Control.Concurrent   ( forkIO, threadDelay )
import System.Posix.Process ( exitImmediately, forkProcess )
import System.Posix.Signals -- ( signalProcess, sigTERM )
import System.Posix.IO	    ( fdToHandle, closeFd, stdError, stdOutput, stdInput
			    , dupTo, createPipe)
import System.IO	    ( hPutStr, hClose, hGetContents, Handle )
import System.Exit	    ( ExitCode(ExitSuccess) )
import System.Cmd	    ( system )
import System.Environment   ( getEnv )

import IO.Process

runCommandWithTimeout :: Maybe Int -> FilePath -> [String] -> IO (Handle, Handle, Handle)
runCommandWithTimeout timeout cmd args =
    do	(p0_in, p1_out)	    <- createPipe
	(p1_in, p0_out)	    <- createPipe
	(p0_err, p1_err)    <- createPipe
	forkProcess $ do mapM_ closeFd [p0_in, p0_out, p0_err]
			 dupTo p1_in stdInput
			 dupTo p1_out stdOutput
			 dupTo p1_err stdError
			 installHandler sigCHLD
				(Catch $ exitImmediately ExitSuccess)
				Nothing
			 p <- runCommand cmd args
			 case timeout of
			    Nothing -> waitForEver
			    Just ms -> do threadDelay (1000 * ms)
					  terminateProcess p
	mapM_ closeFd [p1_in, p1_out, p1_err]
	h_in <- fdToHandle p0_in
	h_out <- fdToHandle p0_out
	h_err <- fdToHandle p0_err
	return (h_in,h_out,h_err)

invoke' :: Maybe Int -> FilePath -> [String] -> String -> IO (String, String)
invoke' timeout cmd args input =
    do	(inn,out,err) <- runCommandWithTimeout timeout cmd args
	hPutStr out input
	hClose out
	i <- hGetContents inn
	e <- hGetContents err
	return (i,e)

invoke :: FilePath -> [String] -> String -> IO (String, String)
invoke = invoke' Nothing

invokeWithTimeOut :: Int -> FilePath -> [String] -> String -> IO (String, String)
invokeWithTimeOut timeout = invoke' (Just timeout)

invoke_ :: FilePath -> [String] -> IO (String, String)
invoke_ cmd args = invoke' Nothing cmd args ""

invokeWithTimeOut_ :: Int -> FilePath -> [String] -> IO (String, String)
invokeWithTimeOut_ timeout cmd args = invoke' (Just timeout) cmd args ""

