
module Plugins.FOL.Primitive where

import Plugins.FOL.Types

-- | > equal = Predicate "equal"
--
--   This refers to the predefined equality in otter.
equal :: [FOLTerm] -> Formula
equal	= Predicate kwEqual
kwEqual = "equal"

-- | > app = FOLFun "app_"
--
--   This is just an arbitrary constant which we hope doesn't
--   conflict with user constants.
app :: [FOLTerm] -> FOLTerm
app	= FOLFun kwApp
kwApp = "app_"

fol_propPairElimination = "__fol_pair_elimination"
propPairElimination =
    (	ForAll "x" $ \x ->
	ForAll "y" $ \y ->
	    equal [ proj [ pair [x, y], FOLFun kwFst [] ], x ]
		`And`
	    equal [ proj [ pair [x, y], FOLFun kwSnd [] ], y ]
    )

fol_propPairInjectivity = "__fol_pair_injectivity"
propPairInjectivity =
    (	ForAll "x" $ \x ->
	ForAll "y" $ \y ->
	ForAll "z" $ \z ->
	ForAll "w" $ \w ->
	    equal [ pair [x,y], pair [z,w] ]
	    `Implies`
	    (equal [x,z] `And` equal [y,w])
    )

fol_primitives =
    [ (fol_propPairInjectivity, propPairInjectivity)
    , (fol_propPairElimination, propPairElimination)
    ]

pair :: [FOLTerm] -> FOLTerm
pair = FOLFun kwPair
kwPair = "pair_"

proj :: [FOLTerm] -> FOLTerm
proj = FOLFun kwProj
kwProj = "proj_"

kwFst = "fst_"
kwSnd = "snd_"

sigma :: [FOLTerm] -> FOLTerm
sigma = FOLFun kwSigma
kwSigma = "sigma_"

fun :: [FOLTerm] -> FOLTerm
fun = FOLFun kwPi
kwPi = "pi_"

