
module TypeChecker.Monad.MetaVars where

import Control.Monad.State
import Control.Monad.Reader

import Syntax.Position
import Syntax.Internal
import Syntax.Values

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.MetaVars.Free
import TypeChecker.Values.Refresh
import TypeChecker.Errors

import Utils.Error
import Utils.Monads
import Utils.Lists

newMetaVariable :: Range -> Term -> Int -> Value -> TC ()
newMetaVariable r x n v =
    unlessM (envAllowUnsolvedMetas <$> ask) $
	modify $ \s -> s { tcMetaVars = MVI r x n v : tcMetaVars s }

getMetaVariables :: TC [MetaVarInfo]
getMetaVariables = tcMetaVars <$> get

allowUnsolvedMetas :: TC a -> TC a
allowUnsolvedMetas = local $ \e -> e { envAllowUnsolvedMetas = True }

clearMetaVariables :: TC ()
clearMetaVariables = modify $ \s -> s { tcMetaVars = [] }

checkThatAllMetasHaveBeenSolved :: TC ()
checkThatAllMetasHaveBeenSolved =
    do	ms <- getMetaVariables
	mapM_ checkClosed ms
    where
	checkClosed (MVI r x n v) =
	    sayWhere r $
	    do	v' <- refresh v
		closed <- isClosed v'
		unless closed $
		    case vAppView v' of
			VAVMeta _ _ ->
			    case x of
				Var _ DontCare	-> failure $ unsolvedUserMetaError r
				_		-> failure $ unsolvedMetaError x n
			_	    ->
			    case x of
				Var _ DontCare	-> failure $ partiallySolvedUserMetaError r v'
				_		-> failure $ partiallySolvedMetaError x n v'

-- | Postpone an action awaiting instantiation of a meta variable
postpone :: MetaRef -> TC () -> TC ()
postpone r action =
    do	env <- ask
	let cl = (env, action)
	ps <- tcPostponedConstraints <$> get
	let ps' = case extract (waitingOn r) ps of
		    Just ((_,cls),ps')  -> (r,cl:cls) : ps'
		    Nothing		-> (r,[cl]) : ps
	modify $ \s -> s { tcPostponedConstraints = ps' }
    where
	waitingOn r (r',_) = r == r'

-- | Wake postponed constraints
wake :: MetaRef -> TC ()
wake r =
    do	ps <- tcPostponedConstraints <$> get
	case extract (waitingOn r) ps of
	    Just ((_,cls),ps')	->
		do  modify $ \s -> s { tcPostponedConstraints = ps' }
		    mapM_ eval cls
	    Nothing		-> return ()
    where
	waitingOn r (r',_) = r == r'
	eval (env,action) = local (const env) action


