
data Bool : Set where
  true  : Bool
  false : Bool

data Nat : Set where
  zero : Nat
  succ : Nat -> Nat

natrec : (n : Nat) ->
	 (P : Nat -> Set) ->
	 P 0 ->
	 ((m : Nat) -> P m -> P (succ m)) ->
	 P n
natrec 0 P z s	      = z
natrec (succ n) P z s = s n (natrec n P z s)

case : (n : Nat) ->
       (P : Nat -> Set) ->
       P 0 -> ((m : Nat) -> P (succ m)) ->
       P n
case n P z s  =
  natrec n {P} of
      0		-> z
      (succ m)	-> \_ -> s m

isZero : Nat -> Bool
isZero n = case n {\_ -> Bool} of
	     0	      -> true
	     (succ m) -> false

isZero' : Nat -> Bool
isZero' n = case n (\_ -> Bool) true (\m -> false)

infixl 10 +
infixl 12 *

(+) : Nat -> Nat -> Nat
x + 0	    = x
x + succ y  = succ (x + y)

(*) : Nat -> Nat -> Nat
x * 0	    = 0
x * succ y  = x * y + x

parity :
       (n : Nat) ->
       (P : Nat -> Set) ->
       ((k : Nat) -> P (2 * k)) ->
       ((k : Nat) -> P (2 * k + 1)) ->
       P n
parity 0 P even odd	    = even 0
parity (succ n) P even odd  =
    parity n {\x -> P (succ x)} of
      (2 * k)	  -> odd k
      (2 * k + 1) -> even (succ k)

data Equals : Set where equals : Equals

data HalfFor (n : Nat) : Set where
  return : Nat -> HalfFor n

halfFor : (n : Nat) -> Equals -> Nat -> HalfFor n
halfFor n equals m = return m

call (n : Nat) : HalfFor n -> Nat
call (return m) = m

half : Nat -> Nat
half n = 
  call (
    parity n {HalfFor} of
	   (2 * k)	-> halfFor (2*k) equals k
	   (2 * k + 1)	-> halfFor (2*k + 1) equals k
  )

compare :
	(n, m : Nat) ->
	(P : Nat -> Nat -> Set) ->
	((x, y : Nat) -> P x (succ y + x)) ->
	((x, y : Nat) -> P (succ y + x) x) ->
	((x : Nat) -> P x x) ->
	P n m
compare 0 0 P lt gt eq		      = eq 0
compare (succ n) 0 P lt gt eq	      = gt 0 n
compare 0 (succ m) P lt gt eq	      = lt 0 m
compare (succ n) (succ m) P lt gt eq  =
    compare n m {\x y -> P (succ x) (succ y)} of
	x	     (succ y + x) -> lt (succ x) y
	(succ y + x) x		  -> gt (succ x) y
	x	     x		  -> eq (succ x)

diff : Nat -> Nat -> Nat
diff n m = compare n m {\_ _ -> Nat} of
	      x		   (succ y + x) -> succ y
	      (succ y + x) x		-> succ y
	      x		   x		-> 0

