-- an example from Michael Beeson

-- # import Property

-- some logical axioms

infix 14 *
infix 10 +
infix 6 ==

Ex : (A:Set) -> (A->Prop) -> Prop
witness : (A:Set) -> (P:A->Prop) -> (x:A) -> P x -> Ex A P
elimEx : (A:Set) -> (P:A->Prop) -> Ex A P -> (C:Prop) -> ((x:A)->P x ->C) -> C

-- natural numbers

data Nat : Set where
 zero : Nat
 succ : Nat -> Nat

indNat : (P:Nat -> Prop) -> P zero -> ((n:Nat) -> P n -> P (succ n)) -> (x:Nat) -> P x
indNat P a g zero = a
indNat P a g (succ n) = g n (indNat P a g n)

-- general definitions

isRing : (X:Set) -> (X->X->X) -> (X->X->X) -> (X->X) -> X -> X -> Prop

isRing R (+) (*) minus Zero One =
 (x:R) -> (y:R) -> (z:R) ->
  (   Eq R (x + y) (y + x)
   /\ Eq R (x + Zero) x
   /\ Eq R (x + (minus x)) Zero
   /\ Eq R (x + (y+z)) ((x+y)+z)
   /\ Eq R (x * (y+z)) ((x*y) + (x*z))
   /\ Eq R ((y+z)* x) ((y*x) + (z*x))
   /\ Eq R (x * One) x
   /\ Eq R (One * x) x
   /\ Eq R (x * (y*z)) ((x*y)*z)
  )

isIntegral : (X:Set) -> (X->X->X) -> (X->X->X) -> (X->X) -> X -> X -> Prop

isIntegral R (+) (*) minus Zero One =
     isRing R (+) (*) minus Zero One 
 /\ ((x:R)->(y:R) -> Eq R (x*y) Zero ==> (Eq R x Zero \/ Eq R y Zero))

-- the example

R : Set
(+) : R -> R -> R
(*) : R -> R -> R
minus : R -> R 
Zero : R
One : R

axR : isIntegral  R (+) (*) minus Zero One 

(==) : R -> R -> Prop
(==) = Eq R


power : Nat -> R -> R
power zero x = One
power (succ n) x = (power n x) * x

isZero : R -> Prop
isZero x = x == Zero

cancel : (x:R) -> (y:R) -> (x + y) == y ==> isZero x
cancel x y h =  
 let
   rem : isZero (x + (y + minus y)) 
   rem = fol-plugin(h,axR)
 in
  fol-plugin(rem,axR)

lemZero : (x:R) -> isZero (x * Zero) 
lemZero x = 
 let
   rem0 : Zero + One == One
   rem0 = fol-plugin(axR)
   rem : x * (Zero + One) == x * Zero + x * One 
   rem = fol-plugin(axR)
   rem1 : x * Zero + x * One == x * One
   rem1 = fol-plugin(axR,rem0,rem)
 in
  fol-plugin(rem1,cancel)
 
lemOneZero : (x:R) -> One == Zero ==> isZero x
lemOneZero x h = fol-plugin(axR,h,lemZero)

prop : R -> Nat -> Prop
prop x n = isZero (power n x) ==> isZero x

lem1 : (x:R) -> (n:Nat) -> prop x n
lem1 x =
 let
  base : prop x zero
  base = fol-plugin(lemOneZero)
  step : (n:Nat) -> prop x n -> prop x (succ n)
  step = \ n -> \ h1 -> \ h2 -> fol-plugin(axR,h1,h2)
 in
  indNat (prop x) base step

isNilpotent : R -> Prop
isNilpotent x = Ex Nat (\ n -> isZero (power n x))

thm : (x:R) -> isNilpotent x -> isZero x
thm x h = elimEx Nat (\ n -> isZero (power n x)) h (isZero x) (lem1 x)

     
